"""Top level cookbook."""
import argparse

__title__ = __doc__


def argument_parser():
    """As required by spicerack._cookbook."""
    return argparse.ArgumentParser("Argparse")


def run(args, _):
    """As required by spicerack._cookbook."""
    print(args)
    return 0



"""Cookbook module."""
import argparse
from abc import ABCMeta, abstractmethod
from typing import Optional

from spicerack import Spicerack

ROLLBACK_FAIL_RETCODE: int = 93
"""Reserved exit code: the cookbook had failed and raised an exception while executing the rollback() method."""
CLASS_FAIL_INIT_RETCODE: int = 94
"""Reserved exit code: failed to initialize the cookbook."""
GET_ARGS_PARSER_FAIL_RETCODE: int = 95
"""Reserved exit code: the call to get the argument parser failed."""
PARSE_ARGS_FAIL_RETCODE: int = 96
"""Reserved exit code: a cookbook failed to parse arguments."""
INTERRUPTED_RETCODE: int = 97
"""Reserved exit code: a cookbook execution was interrupted."""
NOT_FOUND_RETCODE: int = 98
"""Reserved exit code: no cookbook was found for the selection."""
EXCEPTION_RETCODE: int = 99
"""Reserved exit code: a cookbook raised an exception while executing."""


class ArgparseFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    """Custom argparse formatter class for cookbooks.

    It can be used as the ``formatter_class`` for the ``ArgumentParser`` instances and it has the capabilities of
    both :py:class:`argparse.ArgumentDefaultsHelpFormatter` and :py:class:`argparse.RawDescriptionHelpFormatter`.
    """


class CookbookBase(metaclass=ABCMeta):
    """Base Cookbook class that all cookbooks must extend to use the class-based API to define a cookbook.

    The class will be instantiated without any parameter.
    """

    spicerack_path: str
    """Reserved class property used by Spicerack internally to track the Cookbook's path."""
    spicerack_name: str
    """Reserved class property used by Spicerack internally to track the Cookbook's name."""

    def __init__(self, spicerack: Spicerack):
        """Initialize the instance and store the Spicerack instance into ``self.spicerack``.

        Arguments:
            spicerack: the Spicerack accessor instance with which the cookbook can access all the Spicerack
                capabilities.

        """
        self.spicerack = spicerack

    @property
    def title(self) -> str:
        """Returns the title of the Cookbook, must be a single line string.

        The default implementation returns the first line of the class docstring if there is any, a single
        dash otherwise.
        """
        if self.__doc__ is None:
            return "-"

        return self.__doc__.splitlines()[0]

    def argument_parser(self) -> argparse.ArgumentParser:
        """Optionally define an argument parser for the cookbook, if the cookbook accepts CLI arguments.

        The default implementation returns an empty ``ArgumentParser`` instance that doesn't accept any arguments and
        uses the class docstring as description. The arguments will be parsed by the Spicerack framework.
        """
        return argparse.ArgumentParser(description=self.__doc__, formatter_class=ArgparseFormatter)

    @abstractmethod
    def get_runner(self, args: argparse.Namespace) -> "CookbookRunnerBase":
        """Return the runner object that will be used to execute the cookbook.

        Derived classes must override this method and can perform any initialization and validation of the parsed
        arguments, but must not perform any real action here.

        If any exception is raised in this method Spicerack will not execute the cookbook.

        Arguments:
            args: the parsed arguments that were parsed using the defined ``argument_parser()``.

        Raises:
            BaseException: any exception raised in the ``get_runner()`` method will be catched by Spicerack and the
                Cookbook will not be executed.

        Returns:
            Must return an instance of a custom class derived from :py:class:`spicerack.cookbook.CookbookRunnerBase`
            that implements the actual execution of the cookbook.

        """


class CookbookRunnerBase(metaclass=ABCMeta):
    """Base class that all cookbooks must extend to use the class-based API to define the execution plan.

    The constructor of the class is left outside of the interface contract so that each cookbook is free to customize
    it as needed.
    """

    @property
    def runtime_description(self) -> str:
        """Optional message to be used as the runtime description of the cookbook.

        Cookbooks can override this instance property to define their custom description, also based on the given
        command line arguments. For example this will be used in the task start/end messages. The default
        implementation returns an empty string.
        """
        return ""

    @abstractmethod
    def run(self) -> Optional[int]:
        """Execute the cookbook.

        Returns:
            The return code of the cookbook, it should be zero or :py:data:`None` on success, a positive
            integer smaller than ``128`` and not in the range ``90-99``
            (see :ref:`Reserved exit codes<reserved-codes>`) in case of failure.

        Raises:
            BaseException: any exception raised in the ``run()`` method will be catched by Spicerack and the Cookbook
                execution will be considered failed.

        """

    def rollback(self) -> None:
        """Called by Spicerack when the cookbook fails the execution.

        This method by default does nothing. Cookbooks classes that inherit from this one can override it to add their
        own custom actions to perform on error to rollback to a previous state.
        The method will be called if the cookbook raises any un-caught exception or exits with a non-zero exit code.
        For example it can be used to cleanup any left-over inconsistent state as if the cookbook was never run.

        Any un-caught exception raised by this method will be caught by Spicerack and logged, along with the original
        exit code of the cookbook. The final exit code will be the reserved
        :py:const:`cookbooks.ROLLBACK_FAIL_RETCODE`.
        """



"""MediaWiki module."""
import logging
from typing import Any

from cumin.transports import Command
from wmflib.constants import CORE_DATACENTERS
from wmflib.requests import http_session

from spicerack.confctl import ConftoolEntity
from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError
from spicerack.remote import Remote, RemoteExecutionError, RemoteHosts

logger = logging.getLogger(__name__)


class MediaWikiError(SpicerackError):
    """Custom exception class for errors in this module."""


class MediaWikiCheckError(SpicerackCheckError):
    """Custom exception class for checking errors in this module."""


class MediaWiki:
    """Class to manage MediaWiki-specific resources."""

    _siteinfo_url: str = "https://api.svc.{dc}.wmnet/w/api.php?action=query&meta=siteinfo&format=json&formatversion=2"
    """The URL of the siteinfo API to be formatted with a specific ``dc``."""

    _config_file_base_url: str = "https://mw-misc.discovery.wmnet:30443/conf/"
    """The URL of the internal service responding for noc.wikimedia.org/conf"""

    def __init__(self, conftool: ConftoolEntity, remote: Remote, user: str, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            conftool: the conftool instance for the mwconfig type objects.
            remote: the Remote instance.
            user: the name of the effective running user.
            dry_run: whether this is a DRY-RUN.

        """
        self._conftool = conftool
        self._remote = remote
        self._user = user
        self._dry_run = dry_run
        self._http_session = http_session(".".join((self.__module__, self.__class__.__name__)), timeout=10)

    def check_config_line(self, filename: str, expected: str) -> bool:
        """Check that a MediaWiki configuration file contains some value.

        Arguments:
            filename: filename without extension in wmf-config.
            expected: string expected to be found in the configuration file.

        Returns:
            :py:data:`True` if the expected string is found in the configuration file, :py:data:`False` otherwise.

        Raises:
            requests.exceptions.RequestException: on error.

        """
        url = f"{self._config_file_base_url}{filename}.php.txt"
        mwconfig = self._http_session.get(url, headers={"Host": "noc.wikimedia.org"})
        found = expected in mwconfig.text
        logger.debug(
            "Checked message (found=%s) in MediaWiki config %s:\n%s",
            found,
            url,
            expected,
        )

        return found

    def get_siteinfo(self, datacenter: str) -> dict:
        """Get the JSON paylod for siteinfo from a random host in a given datacenter.

        Arguments:
            datacenter: the DC where to query for siteinfo.

        Raises:
            requests.exceptions.RequestException: on failure.

        """
        url = MediaWiki._siteinfo_url.format(dc=datacenter)
        headers = {"Host": "en.wikipedia.org"}

        response = self._http_session.get(url, headers=headers, timeout=3)
        response.raise_for_status()

        return response.json()

    def check_siteinfo(self, datacenter: str, checks: dict[tuple[str, ...], Any], samples: int = 1) -> None:
        """Check that a specific value in siteinfo matches the expected ones, on multiple hosts.

        Arguments:
            datacenter: the DC where to query for siteinfo.
            checks: dictionary of items to check, in which the keys are tuples with the path of keys to traverse
                the siteinfo dictionary to get the value and the values are the expected values to check. To check
                ``siteinfo[key1][key2]`` for a value ``value``, use::

                    {('key1', 'key2'): 'value'}

            samples: the number of different calls to siteinfo to perform.

        Raises:
            spicerack.mediawiki.MediaWikiError: if unable to get siteinfo or unable to traverse the siteinfo
                dictionary after all tries.
            spicerack.mediawiki.MediaWikiCheckError: if the value doesn't match after all tries.

        """
        for i in range(1, samples + 1):  # Randomly check different hosts from the load balancer
            logger.debug("Checking siteinfo %d/%d", i, samples)
            self._check_siteinfo(datacenter, checks)

    def scap_sync_config_file(self, filename: str, message: str) -> None:
        """Execute scap sync-file to deploy a specific configuration file of wmf-config.

        Arguments:
            filename: the filename without extension of wmf-config.
            message: the message to use for the scap sync-file execution.

        Raises:
            spicerack.remote.RemoteExecutionError: on error.

        """
        logger.debug("Syncing MediaWiki wmf-config/%s.php", filename)
        query = "C:Deployment::Rsync and R:Class%cron_ensure = absent"
        command = f"su - {self._user} -c 'scap sync-file --force wmf-config/{filename}.php \"{message}\"'"
        self._remote.query(query).run_sync(command)

    def set_readonly(self, datacenter: str, message: str) -> None:
        """Set the Conftool readonly variable for MediaWiki config in a specific datacenter.

        Arguments:
            datacenter: the DC for which the configuration must be changed.
            message: the readonly message string to set in MediaWiki.

        Raises:
            spicerack.confctl.ConfctlError: on Conftool errors and failed validation.
            spicerack.mediawiki.MediaWikiError: on failed siteinfo validation.

        """
        self._conftool.set_and_verify("val", message, scope=datacenter, name="ReadOnly")
        self._check_siteinfo_dry_run_aware(
            datacenter,
            {
                ("query", "general", "readonly"): True,
                ("query", "general", "readonlyreason"): message,
            },
            samples=10,
        )

    def set_readwrite(self, datacenter: str) -> None:
        """Set the Conftool readonly variable for MediaWiki config to False to make it read-write.

        Arguments:
            datacenter: the DC for which the configuration must be changed.

        Raises:
            spicerack.confctl.ConfctlError: on Conftool errors and failed validation.
            spicerack.mediawiki.MediaWikiError: on failed siteinfo validation.

        """
        self._conftool.set_and_verify("val", False, scope=datacenter, name="ReadOnly")
        self._check_siteinfo_dry_run_aware(datacenter, {("query", "general", "readonly"): False}, samples=10)

    def get_master_datacenter(self) -> str:
        """Return a string representing the primary DC."""
        return next(self._conftool.get(scope="common", name="WMFMasterDatacenter")).val

    def set_master_datacenter(self, datacenter: str) -> None:
        """Set the MediaWiki config master datacenter variable in Conftool.

        Arguments:
            datacenter: the new master datacenter.

        Raises:
            spicerack.confctl.ConfctlError: on error.

        """
        self._conftool.set_and_verify("val", datacenter, scope="common", name="WMFMasterDatacenter")
        for dc in CORE_DATACENTERS:
            self._check_siteinfo_dry_run_aware(
                dc,
                {("query", "general", "wmf-config", "wmfMasterDatacenter"): datacenter},
                samples=10,
            )

    def get_maintenance_host(self, datacenter: str) -> RemoteHosts:
        """Get an instance to execute commands on the maintenance hosts in a given datacenter.

        Arguments:
            datacenter: the datacenter to filter for.

        """
        return self._remote.query("A:mw-maintenance and A:" + datacenter)

    def check_periodic_jobs_enabled(self, datacenter: str) -> None:
        """Check that MediaWiki periodic jobs are enabled in the given DC.

        Arguments:
            datacenter: the name of the datacenter to work on.

        Raises:
            spicerack.remote.RemoteExecutionError: on failure.

        """
        self.get_maintenance_host(datacenter).run_sync(
            # List all timers that start with mediawiki_job_
            "systemctl list-units 'mediawiki_job_*' --no-legend "
            # Just get the timer name
            "| awk '{print $1}' "
            # For each, check `systemd is-enabled`, which will fail if
            # the unit is disabled. xargs will exit with failure if
            # any of the is-enabled checks fail
            "| xargs -n 1 sh -c 'systemctl is-enabled $0'",
            is_safe=True,
        )

    def check_periodic_jobs_disabled(self, datacenter: str) -> None:
        """Check that MediaWiki periodic jobs are not enabled in the given DC.

        Arguments:
            datacenter: the name of the datacenter to work on.

        Raises:
            spicerack.remote.RemoteExecutionError: on failure.

        """
        targets = self.get_maintenance_host(datacenter)
        targets.run_async(
            Command(
                # List all timers that start with mediawiki_job_
                "systemctl list-units 'mediawiki_job_*' --no-legend "
                # Just get the timer name
                "| awk '{print $1}' "
                # For each, check `systemd is-enabled`, which will pass if
                # the unit is enabled. Invert the status code so only disabled
                # pass. 255 instructs xargs to immediately abort.
                "| xargs -n 1 sh -c 'systemctl is-enabled $0 && exit 255 || exit 0'",
            ),
            is_safe=True,
        )

    def stop_periodic_jobs(self, datacenter: str) -> None:
        """Remove and ensure MediaWiki periodic jobs are disabled in the given DC.

        Arguments:
            datacenter: the name of the datacenter to work on.

        Raises:
            spicerack.remote.RemoteExecutionError: on failure.

        """
        targets = self.get_maintenance_host(datacenter)
        logger.info("Disabling MediaWiki periodic jobs in %s", datacenter)

        pkill_ok_codes = [0, 1]  # Accept both matches and no matches
        # Stop all systemd job units and timers
        targets.run_async("systemctl stop mediawiki_job_*")
        targets.run_async(
            # Kill MediaWiki wrappers, in case someone has started one manually. See modules/scap/manifests/scripts.pp
            # in the Puppet repo.
            Command('pkill --full "/usr/local/bin/foreachwiki"', ok_codes=pkill_ok_codes),
            Command(
                'pkill --full "/usr/local/bin/foreachwikiindblist"',
                ok_codes=pkill_ok_codes,
            ),
            Command('pkill --full "/usr/local/bin/expanddblist"', ok_codes=pkill_ok_codes),
            Command('pkill --full "/usr/local/bin/mwscript"', ok_codes=pkill_ok_codes),
            Command('pkill --full "/usr/local/bin/mwscriptwikiset"', ok_codes=pkill_ok_codes),
            # Kill all remaining PHP (but not php-fpm) processes for all users
            Command("killall -r 'php$'", ok_codes=[]),
            "sleep 5",
            # No more time to be gentle
            Command("killall -9 -r 'php$'", ok_codes=[]),
            "sleep 1",
        )
        self.check_periodic_jobs_disabled(datacenter)

        try:
            # Look for remaining PHP (but not php-fpm) processes. php-fpm is used for
            # serving noc.wikimedia.org, which is independent of periodic jobs
            targets.run_sync("! pgrep -c 'php$'", is_safe=True)
        except RemoteExecutionError:
            # We just log an error, don't actually report a failure to the system. We can live with this.
            logger.error("Stray php processes still present on the %s maintenance host, please check", datacenter)

    @retry(
        tries=5,
        backoff_mode="constant",
        exceptions=(MediaWikiError, MediaWikiCheckError),
    )
    def _check_siteinfo(self, datacenter: str, checks: dict[tuple[str], Any]) -> None:
        """Check that a specific value in siteinfo matches the expected ones, retrying if doesn't match.

        Arguments:
            datacenter: the DC where to query for siteinfo.
            checks: dictionary of items to check, in which the keys are tuples with the path of keys to traverse
                the siteinfo dictionary to get the value and the values are the expected values to check. To check
                ``siteinfo[key1][key2]`` for a value ``value``, use::

                    {('key1', 'key2'): 'value'}

        Raises:
            spicerack.mediawiki.MediaWikiError: if unable to get siteinfo or unable to traverse the siteinfo dictionary
                after all tries.
            spicerack.mediawiki.MediaWikiCheckError: if the value doesn't match after all tries.

        """
        try:
            siteinfo = self.get_siteinfo(datacenter)
        except Exception as e:
            raise MediaWikiError("Failed to get siteinfo") from e

        for path, expected in checks.items():
            value = siteinfo.copy()  # No need for deepcopy, it will not be modified
            for key in path:
                try:
                    value = value[key]
                except (KeyError, TypeError) as e:
                    raise MediaWikiError(f"Failed to traverse siteinfo for key {key}") from e

            if value != expected:
                raise MediaWikiCheckError(f"Expected '{expected}', got '{value}' for path: {path}")

    def _check_siteinfo_dry_run_aware(
        self, datacenter: str, checks: dict[tuple[str, ...], Any], samples: int = 1
    ) -> None:
        """Dry-run mode aware check_siteinfo. See check_siteinfo() documentation for more details.

        Arguments:
            datacenter: the DC where to query for siteinfo.
            checks: dictionary of items to check, in which the keys are tuples with the path of keys to traverse
                the siteinfo dictionary to get the value and the values are the expected values to check. To check
                ``siteinfo[key1][key2]`` for a value ``value``, use::

                    {('key1', 'key2'): 'value'}

            samples: the number of different calls to siteinfo to perform.

        Raises:
            spicerack.mediawiki.MediaWikiError: if unable to get siteinfo or unable to traverse the siteinfo dictionary
                after all tries.
            spicerack.mediawiki.MediaWikiCheckError: if the value doesn't match after all tries.

        """
        if self._dry_run:
            logger.debug("Reset samples to check_siteinfo from %s to 1 in dry-run mode", samples)
            samples = 1

        try:
            self.check_siteinfo(datacenter, checks, samples=samples)
        except (MediaWikiError, MediaWikiCheckError) as e:
            if self._dry_run:
                logger.info(e)
            else:
                raise



"""Class API rollback_raise cookbook."""
import argparse

from spicerack.cookbook import CookbookBase, CookbookRunnerBase


class RollbackRaiseCookbook(CookbookBase):
    """Class API rollback_raise cookbook."""

    def argument_parser(self):
        """As defined by the parent class."""
        return argparse.ArgumentParser("Argparse")

    def get_runner(self, args):
        """As required by the parent class."""
        return RollbackRaiseRunner()


class RollbackRaiseRunner(CookbookRunnerBase):
    """As required by spicerack._cookbook."""

    def run(self):
        """As required by the parent class."""
        return 1

    def rollback(self):
        """Rollback raises."""
        raise RuntimeError("rollback has raised")



"""Tests package for Spicerack."""
from pathlib import Path

TESTS_BASE_PATH = Path(__file__).parent.resolve()


def get_fixture_path(*paths):
    """Return the absolute path of the given fixture.

    Arguments:
        *paths: arbitrary positional arguments used to compose the absolute path to the fixture.

    Returns:
        str: the absolute path of the selected fixture.

    """
    return Path(TESTS_BASE_PATH, "fixtures", *paths)


SPICERACK_TEST_PARAMS = {
    "cumin_config": get_fixture_path("remote", "config.yaml"),
    "cumin_installer_config": get_fixture_path("remote", "config_installer.yaml"),
    "conftool_config": get_fixture_path("confctl", "config.yaml"),
    "conftool_schema": get_fixture_path("confctl", "schema.yaml"),
    "debmonitor_config": get_fixture_path("debmonitor", "config.ini"),
    "spicerack_config_dir": get_fixture_path(),
}



"""Package configuration."""

from setuptools import find_packages, setup

with open("README.rst", "r") as readme:
    LONG_DESCRIPTION = readme.read()


INSTALL_REQUIRES = [
    "conftool>=1.0.1",
    "cumin>=3.0.2",
    "dnspython~=2.0.0",
    "elasticsearch>=5.0.0,<7.15.0",
    "elasticsearch-curator~=5.0",
    "gitpython>=3.1.14",
    "kafka-python>=2.0.1",
    "kubernetes==12.0.*",  # frozen to the version available on debian bullseye
    "packaging",
    "pymysql>=0.9.3",
    "pynetbox~=6.6",
    "python-etcd~=0.4.5",
    "redis>=3.5.3,<=4.1.3",
    "requests>=2.25.0",
    "wmflib",
]

# Extra dependencies
EXTRAS_REQUIRE = {
    # Test dependencies
    "tests": [
        "bandit>=1.6.2",
        "flake8>=3.8.4",
        "mypy>=0.812",
        "pytest-cov>=2.10.1",
        "pytest-xdist>=2.2.0",
        "pytest>=6.0.2",
        "requests-mock>=1.7.0",
        # This is required for flake8 to run proprely, as when running tox
        # setuptools comes boundled is usually way older (debian sid has 44 as
        # of writing this).
        "setuptools>=53.0",
        "sphinx_rtd_theme>=1.0",
        "sphinx-argparse>=0.2.5",
        "sphinx-autodoc-typehints>=1.9.0",
        "Sphinx>=3.4.3",
        "types-PyMySQL",
        "types-redis",
        "types-requests",
        "types-setuptools",
    ],
    "format": [
        "black<=21.12b0",  # this is needed so that it doesn't confict with curator
        "isort",
    ],
    "prospector": [
        "prospector[with_everything]>=1.1.7",
        "pytest>=6.0.2",
        "requests-mock>=1.7.0",
    ],
}

SETUP_REQUIRES = [
    "pytest-runner>=2.11.1",
    "setuptools_scm>=5.0.1",
]

setup(
    author="Riccardo Coccioli",
    author_email="rcoccioli@wikimedia.org",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: BSD",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Clustering",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Systems Administration",
    ],
    description="Automation framework for the WMF production infrastructure",
    entry_points={"console_scripts": ["cookbook = spicerack._cookbook:main"]},
    extras_require=EXTRAS_REQUIRE,
    install_requires=INSTALL_REQUIRES,
    keywords=["wmf", "automation", "orchestration"],
    license="GPLv3+",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    name="wikimedia-spicerack",  # Must be the same used for __version__ in __init__.py
    package_data={"spicerack": ["py.typed"]},
    packages=find_packages(exclude=["*.tests", "*.tests.*"]),
    platforms=["GNU/Linux"],
    python_requires=">=3.9",
    setup_requires=SETUP_REQUIRES,
    use_scm_version=True,
    url="https://github.com/wikimedia/operations-software-spicerack",
    zip_safe=False,
)



"""Kafka Module test."""
from collections import namedtuple
from typing import Union
from unittest import mock

import pytest
from kafka import KafkaConsumer, TopicPartition
from kafka.structs import OffsetAndMetadata, OffsetAndTimestamp
from wmflib.config import load_yaml_config

from spicerack.kafka import DELTA, ConsumerDefinition, Kafka, KafkaError
from spicerack.tests import get_fixture_path

OFFSET = 2

TIMESTAMP = 12323847623

SimpleMessage = namedtuple("SimpleMessage", "timestamp offset")

test_data = [
    (
        (
            ["wikidata"],
            ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_1"),
            ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_2"),
        ),
        [({TopicPartition(topic="eqiad.wikidata", partition=0): OffsetAndMetadata(OFFSET, None)})],
    ),
    (
        (
            ["wikidata"],
            ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_1"),
            ConsumerDefinition("eqiad", "main", "consumer_main_2"),
        ),
        [({TopicPartition(topic="eqiad.wikidata", partition=0): OffsetAndMetadata((TIMESTAMP - DELTA) - 100, None)})],
    ),
    (
        (
            ["wikidata", "mediainfo"],
            ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_1"),
            ConsumerDefinition("eqiad", "main", "consumer_main_2"),
        ),
        [
            ({TopicPartition(topic="eqiad.wikidata", partition=0): OffsetAndMetadata((TIMESTAMP - DELTA) - 100, None)}),
            (
                {
                    TopicPartition(topic="eqiad.mediainfo", partition=0): OffsetAndMetadata(
                        (TIMESTAMP - DELTA) - 100, None
                    )
                }
            ),
        ],
    ),
    (
        (
            ["wikidata"],
            ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_1"),
            ConsumerDefinition("codfw", "main", "consumer_main_2"),
        ),
        [({TopicPartition(topic="codfw.wikidata", partition=0): OffsetAndMetadata((TIMESTAMP - DELTA) - 100, None)})],
    ),
]


def _answer_offsets_for_times(timestamps):
    return {tp: OffsetAndTimestamp(ts - 100, ts) for tp, ts in timestamps.items()}


def _answer_partitions_for_topic(topic) -> Union[list[int], None]:
    return [0] if topic.startswith("eqiad.") or topic.startswith("codfw.") else []


@mock.patch("ssl.SSLContext.load_verify_locations")
@mock.patch("spicerack.kafka.KafkaConsumer", autospec=True)
class TestKafka:
    """Test class for Kafka module."""

    def setup_method(self):
        """Set up the module under test."""
        # pylint: disable=attribute-defined-outside-init
        self.kafka = Kafka(kafka_config=load_yaml_config(get_fixture_path("kafka", "config.yaml")), dry_run=False)

    @pytest.mark.parametrize("func_arguments,expected_commit_params", test_data)
    def test_offset_transfer(self, consumer_patch, load_verify_locations_patch, func_arguments, expected_commit_params):
        """It should correctly transfer offsets between consumer groups."""
        to_consumer_mock = self._setup_consumer_mocks(
            consumer_patch, _answer_offsets_for_times, _answer_partitions_for_topic
        )

        self.kafka.transfer_consumer_position(*func_arguments)
        load_verify_locations_patch.assert_called()
        assert to_consumer_mock.commit.call_count == len(expected_commit_params)
        to_consumer_mock.commit.assert_has_calls(
            (mock.call(params) for params in expected_commit_params), any_order=True
        )

    @pytest.mark.parametrize("func_arguments", test_data)
    def test_offset_transfer_dry_run(self, consumer_patch, _, func_arguments):
        """It should read but not transfer offsets between consumer groups."""
        kafka = Kafka(kafka_config=load_yaml_config(get_fixture_path("kafka", "config.yaml")), dry_run=True)
        to_consumer_mock = self._setup_consumer_mocks(
            consumer_patch, _answer_offsets_for_times, _answer_partitions_for_topic
        )
        kafka.transfer_consumer_position(*func_arguments[0])
        assert not to_consumer_mock.commit.called

    def test_no_source_offset(self, consumer_patch, _):
        """It should raise an exception with specific message if no source offset available."""
        TestKafka._setup_empty_consumer_mocks(consumer_patch)
        with pytest.raises(
            expected_exception=KafkaError, match="Offset not found for topic eqiad.wikidata, partition 0."
        ):
            self.kafka.transfer_consumer_position(
                ["wikidata"],
                ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_1"),
                ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_2"),
            )

    def test_no_source_timestamp(self, consumer_patch, _):
        """It should raise an exception with specific message if no source timestamp available."""
        TestKafka._setup_empty_consumer_mocks(consumer_patch)
        with pytest.raises(
            expected_exception=KafkaError, match="Offset not found for topic eqiad.wikidata, partition 0."
        ):
            self.kafka.transfer_consumer_position(
                ["wikidata"],
                ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_1"),
                ConsumerDefinition("eqiad", "main", "consumer_jumbo_2"),
            )

    def test_no_destination_offset(self, consumer_patch, _):
        """It should raise an exception with specific message if no source timestamp available."""
        TestKafka._setup_consumer_mocks(
            consumer_patch, offset_for_times_answer=lambda x: {TopicPartition("eqiad.wikidata", 0): None}
        )
        with pytest.raises(
            expected_exception=KafkaError,
            match=f"Offset not found for topic eqiad.wikidata, partition 0, when seeking by timestamp {TIMESTAMP}.",
        ):
            self.kafka.transfer_consumer_position(
                ["wikidata"],
                ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_1"),
                ConsumerDefinition("eqiad", "main", "consumer_jumbo_2"),
            )

    def test_set_position_by_timestamp(self, consumer_patch, _):
        """It should approximate offset based on source timestamp."""
        consumer_mock = mock.MagicMock(spec_set=KafkaConsumer)
        consumer_mock.offsets_for_times.side_effect = _answer_offsets_for_times
        consumer_mock.partitions_for_topic.side_effect = _answer_partitions_for_topic
        consumer_patch.side_effect = [consumer_mock]
        self.kafka.set_consumer_position_by_timestamp(
            ConsumerDefinition("eqiad", "main", "consumer-group"), {"wikidata": TIMESTAMP, "mediainfo": TIMESTAMP + 10}
        )

        assert consumer_mock.commit.call_count == 2
        consumer_mock.commit.assert_has_calls(
            [
                mock.call(
                    {
                        TopicPartition(topic="eqiad.wikidata", partition=0): OffsetAndMetadata(
                            (TIMESTAMP - DELTA) - 100, None
                        )
                    }
                ),
                mock.call(
                    {
                        TopicPartition(topic="eqiad.mediainfo", partition=0): OffsetAndMetadata(
                            (TIMESTAMP - DELTA + 10) - 100, None
                        )
                    }
                ),
            ]
        )

    def test_no_topic_partitions(self, consumer_patch, _):
        """It should read but not transfer offsets between consumer groups."""
        kafka = Kafka(kafka_config=load_yaml_config(get_fixture_path("kafka", "config.yaml")), dry_run=False)
        self._setup_consumer_mocks(consumer_patch, _answer_offsets_for_times, lambda _: None)

        with pytest.raises(expected_exception=KafkaError, match="Partitions not found for topic eqiad.wikidata."):
            kafka.transfer_consumer_position(
                ["wikidata"],
                ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_1"),
                ConsumerDefinition("eqiad", "jumbo", "consumer_jumbo_2"),
            )

    @staticmethod
    def _setup_empty_consumer_mocks(consumer_patch):
        from_consumer_mock = mock.MagicMock(spec_set=KafkaConsumer)
        to_consumer_mock = mock.MagicMock(spec_set=KafkaConsumer)
        from_consumer_mock.partitions_for_topic.return_value = [0]
        from_consumer_mock.committed.return_value = None
        from_consumer_mock.__next__.return_value = None
        consumer_patch.side_effect = [from_consumer_mock, to_consumer_mock]

    @staticmethod
    def _setup_consumer_mocks(
        consumer_patch,
        offset_for_times_answer=_answer_offsets_for_times,
        answer_for_partition_for_topic=_answer_partitions_for_topic,
    ):
        from_consumer_mock = mock.MagicMock(spec_set=KafkaConsumer)
        from_consumer_mock.partitions_for_topic.side_effect = answer_for_partition_for_topic
        from_consumer_mock.__next__.return_value = SimpleMessage(timestamp=TIMESTAMP, offset=OFFSET)
        from_consumer_mock.committed.return_value = OFFSET
        to_consumer_mock = mock.MagicMock(spec_set=KafkaConsumer)
        to_consumer_mock.partitions_for_topic.side_effect = (
            lambda topic: [0] if topic.startswith("eqiad.") or topic.startswith("codfw.") else []
        )
        to_consumer_mock.offsets_for_times.side_effect = offset_for_times_answer
        consumer_patch.side_effect = [from_consumer_mock, to_consumer_mock]
        return to_consumer_mock
